'************************
'**** AIM_SOLVE******
'************************
'
'  version 1.1 (1/31/11)
'
'  This model procedure takes a linear model with future-dated variables
'  and calls the Anderson-Moore Algorithm (AMA) to create a model with only
'  contemporaneous and lagged variables.  It requires R and the AMA package
'  and must be run from a command line or in a program.
'
'
'  To install AMA
'    1.  start R
'    2.  enter: options(repos=c(CRAN = "ftp://cran.r-project.org/pub/R"))
'    3.  enter: install.packages("AMA")
'
'
'  Syntax:
'
'    Form A:  <input model_name>.aim_solve(options)
'
'    Form B:  <input model_name>.aim_solve(options) arg1 arg2
'
'  Arguments (form B only)
'
'    arg1:  string variable list of names of endogenous variables
'    arg2:  string variable list of names of shock variables
'
'  Options
'
'    parse=<s/a>         Parsing method     default:  form A, s; form B, a
'    modout=<text>       Name of output model   default = aimmod
'    coefpre=<text>      Coefficient prefix in output model  default = c_
'    mat=<no/yes/only>   Option for retaining AMA matrices in workspace default = no
'    z=<1/2>             Option for how may variables to include in output model equations 
'                          default = 1
'    addshks             Add shock equations (form B) 
'    debug               Do not delete temporary matrices


' ******************************************************************
' ******************************************************************
' ******************************************************************
' ******************************************************************
' Examine the object

  'check that an object exists
  %type =@getthistype
  if %type="NONE" then
    @uiprompt("Error:  No object found, please open a MODEL object.")
    stop
    endif

  'check that _this is a model object
  if %type<>"MODEL" then
    @uiprompt("Error:  AIM_SOLVE can only be run from a MODEL object.")
    stop
    endif



' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 1: Examine arguments
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************

' Two arguments => %procform = B

  %procform = "a"

  if @len(%args) > 0 then
    !nargs = @wcount(%args)
    if !nargs = 2 then
      %procform = "b"
      %varendog = %0  
      if @isobject(%varendog) = 0 then
        %errstring = "Error:  aim_solve argument 1 is not the name of an existing object"
        @uiprompt(%errstring)
        stop
        endif
      if {%varendog}.@type <> "STRING" then
        @uiprompt("Error:  aim_solve argument 1 must be a STRING object.")
        stop
        endif
      %varendog = @lower({%varendog})
      %varshock = %1  
      if @isobject(%varshock) = 0 then
        %errstring = "Error:  aim_solve argument 1 is not the name of an existing object"
        @uiprompt(%errstring)
        stop
        endif
      if {%varshock}.@type <> "STRING" then
        @uiprompt("Error:  aim_solve argument 1 must be a STRING object.")
        stop
        endif
      %varshock = @lower({%varshock})
      else
      %errstring = "Error:  Form B of the aim_solve proc requires two arguments"
      @uiprompt(%errstring)
      stop
      endif
    endif




' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 2: Examine options
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************


' set defaults
  %modout = "aimmod"
  %coefpre = "c_"
  %matrices = "no"
  %debug = "no"
  %bq = "no"
  !z = 1
  %addshks = "no"
  if %procform = "a" then
    %parse = "sim"
    else
    %parse = "arithmetic"
    endif


' set overrides
  if @len(@option(1)) > 0 then
    %temp= @equaloption("coefpre") 
    if @len(%temp)>0 then
      %coefpre = @lower(%temp)
      endif
    %temp= @equaloption("modout")
    if @len(%temp)>0 then
      %modout = %temp
      endif
    %temp= @equaloption("mat")
    if @len(%temp)>0 then
      %matrices = @lower(%temp)
      endif
    if @hasoption("debug") then
      %debug = "yes"
      endif
    %temp= @equaloption("z")
    if @len(%temp)>0 then
      !z = @val(%temp)
      endif
    %temp= @equaloption("parse")
    if @len(%temp)>0 then
      if @lower(%temp) = "s" then
        %parse = "sim"
        endif
      if @lower(%temp) = "a" then
        %parse = "arithmetic"
        endif
      endif
    if @hasoption("addshks") then
      %addshks = "yes"
      endif
    endif


' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 3: Setup
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************

  statusline Setup


' form B proc
  if %procform = "b" and %addshks = "yes" then
  ' add shock equations
    !nshk = @wcount(%varshock)
    for !i = 1 to !nshk
      %t1 = @word(%varshock,!i)
      %eqn = %t1 + "=0"
      _this.append {%eqn}
      next
    endif

' create text view, equation view, and svector form of model
  %modtext=@getnextname("work_")
  %worknames = %worknames + "modtext "
  freeze({%modtext}) _this.text
  %modeqs=@getnextname("work_")
  %worknames = %worknames + "modtext "
  freeze({%modeqs}) _this.equations
  %modsvec = @getnextname("work_")
  %worknames = %worknames + "modsvec "
  svector {%modsvec} = {%modtext}.@svectornb

' form A proc
  if %procform = "a" then
  ' create list of endogenous variables in the same order as their equations appear in the model
    !nvars=@wcount(_this.@varlist)
    !neqs=@wcount(_this.@endoglist)
    %vnames = " "
    for !i = 1 to !neqs
      %t1 = {%modeqs}(!i,3)
      !k1 = @instr(%t1,":")
      !k2 = @instr(%t1,"=")
      if !k2 = 0 then
        !k2 = @len(%t1) + 1
        endif
      %t2 = @mid(%t1,!k1+1,!k2-!k1-1)
      %t2 = @lower(@trim(%t2))
      %vnames = %vnames + " " + %t2
      next
    endif

' form B proc
  if %procform = "b" then
    %vnames = %varendog + " " + %varshock
    !nvars = @wcount(%vnames)
    !neqs = @rows({%modsvec})
    !nendog = @wcount(%varendog)
    !nshk = @wcount(%varshock)
    endif

  if !nvars<>!neqs then
    %errmsg = "Error:  Number of equations (" + @str(!neqs) + ") "
    %errmsg = %errmsg + "does not equal the number of variables (" + @str(!nvars) + ")"
    @uiprompt(%errmsg)
    stop
    endif


  %modname =_this.@name

  !maxlead=0
  !maxlag=0


' output text file
  if @isobject("aim_solve_text") then
    delete aim_solve_text
    endif
  text aim_solve_text
  aim_solve_text.append Input model: name = %modname; form = %procform; parsing = %parse
  if %procform = "b" then
    if %addshks = "yes" then
      aim_solve_text.append Input model:  Shock equations added for: %varshock
      endif
    aim_solve_text.append Input model:  Number of endogenous variables = !nendog
    aim_solve_text.append Input model:  Number of shock variables = !nshk
    endif
  aim_solve_text.append Input model: Number of equations (including any added shock eqs) = !neqs
  if %matrices = "only" then
    aim_solve_text.append Output model: only matrix form created
    else
    if %procform = "a" then
      %outtype = "structural"
      else
      %outtype = "reduced form + structural shocks"
      endif
    aim_solve_text.append Output model: name = %modout; type = %outtype; coef prefix = %coefpre
    endif

' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 4: Simulation-based approach to converting model to matrix form
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' a. Create list of coefficient names
' b. Calculate !maxlead and !maxlag; replace leads with exog vars, so that EViews
'    can perform static sims
' c. Static sims

if %parse = "sim" then


' *************************************************************
' ***************************************
' step a

  %modskeleton = @getnextname("work_")
  %worknames = %worknames + "modskeleton "
  string {%modskeleton} =" "
  %coef_list = @getnextname("work_")
  %worknames = %worknames + "coef_list "
  string {%coef_list} = " "

  for !i1=1 to !neqs
    %temp0 = {%modskeleton} +  @lower({%modsvec}(!i1)) + " "
    {%modskeleton} = %temp0
    if @len(%temp0) > 24000 or !i1 = !neqs then
      call find_coeffs
      {%coef_list} = {%coef_list} + " " + {%modskeleton}
      {%modskeleton} = " "
      endif
    next
  {%coef_list} = @wunique({%coef_list})


' *************************************************************
' ***************************************
' step b

  %ok_chars = "=-+*( "

  for !i2 = 1 to !neqs
    %tmp2 = @lower({%modsvec}(!i2))
    for !i1 = 1 to !nvars
        %tmp1 = @word(%vnames,!i1) + "("
    	!kk = @instr(%tmp2,%tmp1)
    	if !kk>0 then
	    !occurrence = 1
	    while !kk > 0
                'three possibilities
                '  1. it's part of another variable name
                '  2. it's a lag
                '  3. it's a lead
                '
                'look at character in %tmp2 just before %tmp1 to make sure that %tmp1
                'is not part of a longer variable or coefficient name 
	        %before = @mid(%tmp2, !kk-1,1)
                if @instr(%ok_chars,%before) > 0 then
		    %tmp3 = @mid(%tmp2,!kk)
 	     	    !kkll = @instr(%tmp3,"(") 
          	    !kkrr = @instr(%tmp3,")") 
          	    %laglead = @mid(%tmp3,!kkll, !kkrr-!kkll+1)
          	    !ll = 0 + {%laglead}
        	    if !maxlead<!ll then 
		        !maxlead=!ll 'update maxlead
        		endif
	   	    if !maxlag < (-1)*!ll then 
			!maxlag=(-1)*!ll 'update maxlag
        		endif
		    if !ll > 0 then   'it's a lead -- define a new variable 
            		%aaa = @word(%vnames,!i1) + %laglead
            		%bbb = @word(%vnames,!i1) + "_lead_" + @str(!ll)
            		%tmp4 = @replace(%tmp2,%aaa,%bbb,1)
            		%tmp2 = %tmp4
                        'if it's a lead, don't increase the occurrence counter
         		!occurrence = !occurrence - 1
		  	endif
                    endif
      		!occurrence = !occurrence + 1
		!kk = @instr(%tmp2,%tmp1,!occurrence)
		wend
	    endif
	next
    {%modsvec}(!i2) = %tmp2
    next

  if !maxlead = 0 then
    @uiprompt("Error:  The model does not have any leads")
    stop
    endif

  %modnew=@getnextname("work_")
  %worknames = %worknames + "modnew "
  model {%modnew}
  for !i1 = 1 to !neqs
    %zzz = {%modsvec}(!i1)
    {%modnew}.append {%zzz}
    next


' *************************************************************
' ***************************************
' step c:

  statusline Constructing coefficient matrices of EViews model via static sims


' ***************************************
' create new workfile page and copy into it the model and its coefficients
  %mainpagename = @pagename
  pagecopy {%modnew}
  %altpagename = @pagename

  pageselect {%mainpagename}
  for !i = 1 to @wcount({%coef_list})
    %tmp = @word({%coef_list},!i)
    copy {%mainpagename}\{%tmp} {%altpagename}\
    next

  pageselect {%altpagename}

  %endogvars = %vnames
  %exogvars = @lower({%modnew}.@exoglist)
  !perturb = .01
  !nendog = @wcount(%endogvars)
  !nexog = @wcount(%exogvars)


' ***************************************
' data

  smpl @all
  !nobs = @obssmpl

  for !i = 1 to @wcount(%endogvars)
    %tmpname_0 = @word(%endogvars, !i)
    series {%tmpname_0} = 0
    next

  for !i = 1 to @wcount(%exogvars)
    %tmpname_1 = @word(%exogvars, !i)
    series {%tmpname_1} = 0
    next
	

' ***************************************
' static simulations

  %sufbase = "_0"
  %sufmult = "_1"
  %endog_base = @wcross(%endogvars,%sufbase)
  %endog_mult = @wcross(%endogvars,%sufmult)
  %pagestart = @word(@pagesmpl,1)
  call dateshift(%pagestart,%simstart,!maxlag)
  call dateshift(%simstart,%simend,!maxlag)

' create matrices to hold contemporaneous and lag coefficients
  for !i = 0 to !maxlag
    %lagmat_{!i} = @getnextname("workx_")
    matrix {%lagmat_{!i}} = @filledmatrix(!nendog,!nendog,0)
    %worknames = %worknames + "lagmat_" + @str(!i) + " "
    next

' create matrices to hold lead coefficients
  for !i = 1 to !maxlead
    %leadmat_{!i} = @getnextname("workx_")
    matrix {%leadmat_{!i}} = @filledmatrix(!nendog,!nendog,0)
    %worknames = %worknames + "leadmat_" + @str(!i) + " "
    next


' baseline simulation
  smpl %simstart %simend
  {%modnew}.scenario(n,a={%sufbase}) "base"
  {%modnew}.solve(o=b,g=12,z=1e-12)
  %endog_bsols = @getnextname("workx_")
  %worknames = %worknames + "endog_bsols "
  group {%endog_bsols} {%endog_base}
  for !i = 0 to !maxlag
    smpl %simstart + !i %simstart + !i
    %bbname = %endog_bsols + "_" + @str(!i)
    stom({%endog_bsols},{%bbname})
    next

  {%modnew}.scenario(n,a={%sufmult}) "mult"
  %dmat=@getnextname("workx_")
  %worknames = %worknames + "dmat "
  matrix(!nendog,1) {%dmat}


' simulations to identify lead structure of model
  !switch = 0
  smpl %simstart %simstart
  for !i = 1 to !nendog
    for !j = 1 to !maxlead
      %name = @word(%endogvars,!i) + "_lead_" + @str(!j) 
      %intersect = @wintersect(%name,%exogvars)
      if @isempty(%intersect) <> 1 then
        {%name} = {%name} + !perturb
        {%modnew}.solve(d=f)
        {%name} = {%name} - !perturb
        if !switch = 0 then
          %endog_msols = @getnextname("workx_")
          %worknames = %worknames + "endog_msols "
          group {%endog_msols} {%endog_mult}
          !switch = !switch + 1
          endif
        %mmname = %endog_msols + "_0"
        %bbname = %endog_bsols + "_0"
        stom({%endog_msols},{%mmname})
        {%dmat} = ({%mmname} - {%bbname})/!perturb
        matplace({%leadmat_{!j}},{%dmat},!i,1)
        endif
      next
    next


' simulations to identify lag structure of model
  for !i = 1 to !nendog
    %name = @lower(@word(%endogvars,!i))
    smpl %simstart %simstart
    {%name} = {%name} + !perturb
    smpl %simstart %simend
    {%modnew}.solve(d=f)
    smpl %simstart %simstart
    {%name} = {%name} - !perturb
    for !j = 0 to !maxlag
      smpl %simstart + !j %simstart + !j
      %mmname = %endog_msols + "_" + @str(!j)
      %bbname = %endog_bsols + "_" + @str(!j)
      stom({%endog_msols},{%mmname})
      matrix {%dmat} = ({%mmname} - {%bbname})/!perturb
      matplace({%lagmat_{!j}},{%dmat},!i,1)
      next
    next


' ***************************************
' transpose lag and lead matrices so that each equation corresponds to a row,
' and copy matrices back to main workfile page
  for !i = 0 to !maxlag
    {%lagmat_{!i}} = @transpose({%lagmat_{!i}})
    copy {%lagmat_{!i}} {%mainpagename}\
    next
  for !i = 1 to !maxlead
    {%leadmat_{!i}} = @transpose({%leadmat_{!i}})
    copy {%leadmat_{!i}} {%mainpagename}\
    next

' return to main page
  pageselect {%mainpagename}
  pagedelete {%altpagename}

' subtract the identity matrix from the contemporaneous coefficient matrix
  {%lagmat_0} = {%lagmat_0} - @identity(!nendog)

' combine the lead and lag matrices to form hmat
  !n=!maxlag+!maxlead +1
  %hmat=@getnextname("work_")
  %worknames = %worknames + "hmat "
  matrix (!neqs, !n*!neqs){%hmat}
  for !i=(-1)*!maxlag to 0
    !k = (-1)*!i
    matplace({%hmat}, {%lagmat_{!k}}, 1, ((!maxlag-!k)*!neqs +1))
    next
  for !j=1 to !maxlead
    matplace({%hmat}, {%leadmat_{!j}}, 1, ((!maxlag+!j)*!neqs +1))
    next

  endif


' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 5: Arithmetic-based approach to converting model to matrix form
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' a. Calculate !maxlead and !maxlag; replace lags with _lg and leads with _ld notation; 
'    surround all variables with spaces so they are "words".
' b. Treat each equation string as an arithmetic expression (replacing variables with
'    1 or 0) whose value is a coefficient on the variable replaced with 1.
'    each variable

if %parse = "arithmetic" then

  %varx = " "
  %modtextx = @getnextname("work_")
  text {%modtextx}
  %worknames = %worknames + "modtextx "
  %ok_before_chars = "=-+*( "
  %ok_after_chars  = "=-+*(/) "

' delete any blank lines from text view of model
  !nlines = {%modtext}.@linecount
  !nsvec = @rows({%modsvec})
  if !nlines <> !nsvec then
    %modtexta = @getnextname("work_")
    text {%modtexta}
    %worknames = %worknames + "modtexta "
    for !i = 1 to !nlines
      %t1 = {%modtext}.@line(!i)
      if @isempty(%t1) = 0 then
        {%modtexta}.append {%t1}
        endif
      next
    copy {%modtexta} {%modtext}
    endif

' *************************************************************
' ***************************************
' step a
' 
  for !i = 1 to !neqs
    %t1 = @lower({%modtext}.@line(!i))
    %t1 = @replace(%t1," ","")
  ' determine which endog vars appear in the equation
    %z1 = %t1
    call replace_operators
    %t2 = @wintersect(%vnames,%z1)
  ' determine if any lags or leads of endog vars appear
    for !j = 1 to @wcount(%t2)
      %t3 = @word(%t2,!j)
      !kk = @instr(%t1,%t3)
      if !kk > 0 then
	!occurrence = 1
	while !kk > 0
        ' three possibilities
        ' 1. it's part of another variable or coefficient name
        ' 2. it's a contemporaneous, lag, or lead variable
        '
        ' look at character in %t1 just before %t3 and check that %t3
        ' is not part of a longer variable or coefficient name 
	  %before = @mid(%t1, !kk-1,1)
          if @instr(%ok_before_chars,%before) > 0 or !kk = 1 then
	    %t4 = @mid(%t1,!kk)
            !mm = @len(%t3)
            !nn = @len(%t1)
            %after = @mid(%t4,!mm+1,1)
            if @instr(%ok_after_chars,%after) > 0 or (!kk+!mm) > !nn then
              if %after = "(" then
              ' its a lag or lead or contemporaneous in (0) form
   	        !kkll = @instr(%t4,"(") 
                !kkrr = @instr(%t4,")") 
                %laglead = @mid(%t4,!kkll, !kkrr-!kkll+1)
                !ll = 0 + {%laglead}
                if !maxlead < !ll then
	          !maxlead = !ll 'update maxlead
       	          endif
	        if !maxlag < (-1)*!ll then 
	          !maxlag=(-1)*!ll 'update maxlag
                  endif
                %aaa = @word(%t2,!j) + %laglead
                if !ll > 0 then   'it's a lead -- define a new _ld variable 
                  %bbb = " " + @word(%t2,!j) + "_ld_" + @str(!ll) + " "
                  %varx = %varx + " " + %bbb
                  endif
                if !ll <= 0 then   'it's a lag or contemp -- define a new _lg variable 
                  %bbb = " " + @word(%t2,!j) + "_lg_" + @str(-!ll) + " "
                  %varx = %varx + " " + %bbb
                  endif
                %t5 = @left(%t1,!kk-1) + %bbb + @mid(%t1,!kk+!kkrr)
                %t1 = %t5
                !occurrence = !occurrence - 1
                else
              ' its a contemporaneous variable -- define a new _lg_0 variable
                %aaa = @word(%t2,!j)
                %bbb = " " + @word(%t2,!j) + "_lg_0 "
                %varx = %varx + " " + %bbb
                %t5 = @left(%t1,!kk-1) + %bbb + @mid(%t1,!kk+!mm)
                %t1 = %t5
                endif
              endif
            endif
  	  !occurrence = !occurrence + 1
	  !kk = @instr(%t1,%t3,!occurrence)
	  wend
	endif
      next
    {%modtextx}.append {%t1}
    next

  %varx = @wunique(%varx)

' *************************************************************
' ***************************************
' step b: 

  %hmat=@getnextname("work_")
  %worknames = %worknames + "hmat "
  matrix(!neqs,!neqs*(!maxlag + !maxlead + 1)) {%hmat}

' define vary as the list of all variables in the order they appear in hmat
  %vary = " "
  for !i = -!maxlag to 0
    %t1 = "_lg_" + @str(-!i)
    %t2 = @wcross(%vnames,%t1)
    %vary = %vary + " " + %t2
    next
  if !maxlead > 0 then
    for !i = 1 to !maxlead
      %t1 = "_ld_" + @str(!i)
      %t2 = @wcross(%vnames,%t1)
      %vary = %vary + " " + %t2
      next
    endif

  if @wcount(%vary) <> !neqs*(!maxlag + !maxlead + 1) then
    %errmsg = "number of variables at all lags and leads"
    %errmsg = %errmsg + " does not equal the number of columns of hmat"
    @uiprompt(%errmsg)
    stop
    endif

  for !i = 1 to !neqs
    %t1 = " " + {%modtextx}.@line(!i)
  ' remove "=" sign and multiply all terms to its right by minus 1
    !k = @instr(%t1,"=")
    if !k > 0 then
      %t1a = @replace(%t1,"=","-( ")
      %t1 = %t1a + " )"
      endif
  ' determine which vars appear in the equation (lags and leads are treated
  ' as separate variables); in the equation string, replace all variables
  ' except one with "0", replace the remaining variable with "1", calculate
  ' the numerical value of the string, and fill the relelvant element of 
  ' hmat.
    %z1 = %t1
    call replace_operators
    %t2 = @wintersect(%varx,%z1)
    for !j = 1 to @wcount(%t2)
      %t3 = @word(%t2,!j)
      !col = @wfind(%vary,%t3)
      %t3  = " " + %t3
      %t4 = @replace(%t1,%t3,"1")
      for !k = 1 to @wcount(%t2)
        if !k <> !j then
          %t5 = " " + @word(%t2,!k)
          %t6 = @replace(%t4,%t5,"0")
          %t4 = %t6
          endif
        next
      {%hmat}(!i,!col) = -({%t4})
      next
    next      

  endif


  aim_solve_text.append Maximum lag = !maxlag
  aim_solve_text.append Maximum lead = !maxlead

	
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 6:  R step
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' Call AIM (AMA) to compute structural-form (SCOF) and
' reduced-form (BMAT) coefficient matrices

  statusline R step

  xopen(type=r, case=lower) 
  xrun "library('AMA')"

  %neq=@getnextname("work_")
  %worknames = %worknames + "neq "
  scalar {%neq} = !neqs
  %maxleads=@getnextname("work_")
  %worknames = %worknames + "maxleads "
  scalar {%maxleads} = !maxlead
  %maxlags=@getnextname("work_")
  %worknames = %worknames + "maxlags "
  scalar {%maxlags} = !maxlag
  xput {%neq} {%maxleads} {%maxlags} {%hmat}

' call AMA
  %cmd= "res <- callAMA("+@lower(%hmat+","+%neq+","+%maxleads+","+%maxlags)+")"
  xrun {%cmd}

' check for errors
  %errormsg=@getnextname("work_")
  %worknames = %worknames + "errormsg "
  %cmd = %errormsg+"= getReturnCode(res)"
  xrun {%cmd}
  xget {%errormsg}

  %errorcode=@getnextname("work_")
  %worknames = %worknames + "errorcode "
  %cmd = %errorcode+"<- res[9]"
  xrun {%cmd}
  xget(type=scalar,name={%errorcode}) {%errorcode}

  aim_solve_text.append Error code = {{%errorcode}}; {{%errormsg}}

  if {{%errorcode}} <> 0 then
    %errmsg = "AMA Error:  " + {%errormsg} + ".  Execution stopping"
    @uiprompt(%errmsg)
    stop
    endif

' primary method of calculating reduced form (bmat)
  if %bq = "no" then
    %bmat=@getnextname("work_")
    %worknames = %worknames + "bmat "
    '%cmd=%bmat+"<-genBmat("+@lower(%hmat+","+%neq+","+%maxleads+","+%maxlags)+")"
    %cmd=%bmat+"<-array(res[10][[1]],dim=c("+@lower(%neq+","+%neq+"*"+%maxlags)+"))" 
    xrun {%cmd}
    xget {%bmat}
    endif

' alternative method of calculating reduced form (from qmat)
  if %bq = "yes" then
    %cmd=@lower("qrows<-"+%neq+"*"+%maxleads)
    xrun {%cmd}
    %cmd=@lower("qcols<-"+%neq+"*("+%maxlags +"+"+%maxleads+")")
    xrun {%cmd}
    %cmd=@lower("bcols<-"+%neq+"*"+%maxlags)
    xrun {%cmd}
    xrun left <- 1:(qcols-qrows)
    xrun right <- (qcols-qrows+1):qcols
    xrun qmat <- array(res[11][[1]],dim=c(qrows,qcols))
    xrun q_copy <- qmat
    xrun q_copy[,left] <- (-(solve(qmat[,right]))%*% qmat[,left])
    %bmat=@getnextname("work_")
    %worknames = %worknames + "bmat "
    %cmd= %bmat+"<-q_copy[1:"+@lower(%neq)+",1:bcols]"
    xrun {%cmd}
    xget {%bmat}
    aim_solve_text.append reduced form (bmat) calculated from qmat
    endif

' calculate scof
  %scof=@getnextname("work_")
  %worknames = %worknames + "scof "
  %cmd=%scof+"<-genScof("+@lower(%hmat)+","+@upper(%bmat)+","+@lower(%neq)+","+@lower(%maxleads)+","+@lower(%maxlags)+")"
  xrun {%cmd}
  xget {%scof}

  xclose

' *************************************************************************************
' *************************************************************************************
' *************************************************************************************
' SECTION 7:  Translate AIM coeff matrix to EViews model
' *************************************************************************************
' *************************************************************************************
' *************************************************************************************

  if %matrices = "yes" or %matrices = "only" then
    matrix bmat = {%bmat}
    matrix scof = {%scof}
    string varlist = %vnames
    aim_solve_text.append Workspace contains bmat matrix, scof matrix, and varlist string
    endif

  if %matrices <> "only" then


  statusline Creating EViews output model from AMA coefficient matrices

  model {%modout}

  !n = !maxlag + 1
  !n1 = !maxlag*!neqs
  !p = !neqs*!n


  if %procform = "a" then
  ' output model is y = A0*y + A1*y(-1) + ...
    !nn = !neqs
    %cof = @getnextname("work_")
    matrix {%cof} = {%scof}
    %worknames = %worknames + "cof "
  ' adjust %cof because dependent variable will be on equation lhs, not rhs
    for !i=1 to !neqs
      {%cof}(!i,!neqs*!maxlag + !i) = {%cof}(!i,!neqs*!maxlag + !i) + 1
      next
    endif

  if %procform = "b" then
  ' output model is y = (1/(I-A0))*(A1*y(-1) + ... + B*x)
    !nn = @wcount(%varendog)
  ' exclude shock equations from output model => %cof contains just the first !nn rows of scof; 
  ' set the rhs contemporaneous endog coefficients to zero;
  ' multiply the result by the inverse of the matrix of contemporaneous endog coefficients
    %cof = @getnextname("work_")
    matrix(!nn,!n1+!neqs) {%cof}
    %worknames = %worknames + "cof "
    matrix {%cof} = @subextract({%scof},1,1,!nn,!n1+!neqs)
    %zeros = @getnextname("work_")
    matrix(!nn,!nn) {%zeros}
    matplace({%cof},{%zeros},1,!n1+1)
    {%cof} = -@inverse(@subextract({%scof},1,!n1+1,!nn,!n1+!nn))*{%cof}
    endif    
   
' ordered list of variable/lag pairs (matches column order of %cof)
  %allvars=@getnextname("work_")
  string {%allvars} = " "
  %worknames = %worknames + "allvars "
  %lagnames = " "
  for !i = 1 to !maxlag + 1
    !index_lag =  !i - (!maxlag + 1)
    %lagnames = %lagnames + " (" + @str(!index_lag) + ")"
    next
  for !i = 1 to !neqs
    {%allvars} = {%allvars} + " " + @wcross(%vnames,@word(%lagnames,!i))
    next
  {%allvars} = @replace({%allvars},"(0)","")

' eliminate variable/lag pairs that do not appear in any equation
  %colabsmax=@getnextname("work_")
  %worknames = %worknames + "colabsmax "
  vector {%colabsmax} = @cmax(@abs({%cof}))
  %minvars=@getnextname("work_")
  string {%minvars}=""
  %worknames = %worknames + "minvars "
  %mincof = @getnextname("work_")
  matrix {%mincof} = {%cof}
  %worknames = %worknames + "mincof "
  !nonzerocount = 0
  for !j=1 to @columns({%cof})
    if ({%colabsmax}(!j)) > 0 then 
      !nonzerocount=!nonzerocount+1
      {%minvars} =  {%minvars} + " " + @word({%allvars},!j) 
      colplace({%mincof}, @columnextract({%cof}, !j), !nonzerocount)
      endif
    next
  {%mincof} = @subextract({%mincof},1,1,!nn,!nonzerocount)



' ****************************************************************************
' equation option 1:  in each equation, include only variables with non-zero
'                     coefficients and compress each coefficient vector

  if !z = 1 then
    aim_solve_text.append Output model:  each equation contains only variables with non-zero coefficients
    for !j=1 to !nn
      !count = 0
      %lhsvar = @lower(@word(%vnames,!j))
      %coef_ = %coefpre + %lhsvar
      coef(!nonzerocount) {%coef_} = @rowextract({%mincof},!j)
      %eq = " "
      for !k = 1 to !nonzerocount
        !index = !nonzerocount - !k + 1
        !vvv = {%mincof}(!j,!index) 
        if @abs(!vvv) > 1e-12 then
          !count = !count + 1
          {%coef_}(!count) = !vvv
          %eq = %eq + "+" + %coef_ + "("+@str(!count)+")" + "*" + @word({%minvars},!index)
          endif
	next
      %eq = @trim(%eq)
      if @isempty(%eq) then
        %eq = %lhsvar + "=0"
        else
        %eq = %lhsvar + "=" + %eq
        endif
      {%modout}.append %eq
      {%coef_} = @subextract({%coef_},1,1,1,!count)
      next
    endif

' ****************************************************************************
' equation option 2:  in each equation, include all variables with non-zero
'                     coefficients in any equation
  if !z = 2 then
    aim_solve_text.append Output model:  each equation contains all variables with non-zero coefs in any eq
    for !j=1 to !nn
      %lhsvar = @lower(@word(%vnames,!j))
      %coef_ = %coefpre + %lhsvar
      coef(!nonzerocount) {%coef_} = @rowextract({%mincof},!j)
      %eq = " "
      for !k = 1 to !nonzerocount
        !index = !nonzerocount - !k + 1
        !vvv = {%coef_}(!index) 
        if @abs(!vvv) > 1e-12 then
          %eq = %eq + "+" + %coef_ + "("+@str(!index)+")" + "*" + @word({%minvars},!index)
          endif
	next
      %eq = @trim(%eq)
      if @isempty(%eq) then
        %eq = %lhsvar + "=0"
        else
        %eq = %lhsvar + "=" + %eq
        endif
      {%modout}.append %eq
      next
    endif

  endif


' **************************************************************************
' output text and tables

  show aim_solve_text

  if @isobject("aim_solve_work") then
    delete aim_solve_work
    endif
  table(10,2) aim_solve_work

  for !i = 1 to @wcount(%worknames)
    %abc = @word(%worknames,!i)
    aim_solve_work(!i,1) = %abc
    %bcd = "%" + %abc
    aim_solve_work(!i,2) = {%bcd}
    next

  if %debug = "yes" then
    show aim_solve_work
    else
    for !i = 1 to @wcount(%worknames)
      %abc = @word(%worknames,!i)
      %bcd = "%" + %abc
      delete(noerr) {%bcd}
      next
    endif
 



'************************************************************
'***********************************************************
'************************************************************
' subroutines
'************************************************************
'***********************************************************
'************************************************************
subroutine dateshift(string %indate, string %outdate, scalar qtrshift)

' this subroutine takes the quarterly date string in %indate (ie, "2001q1") and shifts 
' it qtrshift quarters, returning the result in %outdate

  !dddd1   = @dateval(%indate,"yyyyfq")
  !dddd2   = @dateadd(!dddd1,qtrshift,"q")
  %outdate = @datestr(!dddd2,"yyyyfq")  

  endsub
  

'************************************************************
'***********************************************************
'************************************************************
subroutine find_coeffs

' operates on the string whose name is in %modskeleton

' replace all delimiters with spaces
  {%modskeleton} = @replace({%modskeleton}, "+", " ")
  {%modskeleton} = @replace({%modskeleton}, "-", " ")
  {%modskeleton} = @replace({%modskeleton}, "*", " ")
  {%modskeleton} = @replace({%modskeleton}, "/", " ")
  {%modskeleton} = @replace({%modskeleton}, "^", " ")
  {%modskeleton} = @replace({%modskeleton}, "=", " ")
  {%modskeleton} = @replace({%modskeleton}, "(", " ")
  {%modskeleton} = @replace({%modskeleton}, ")", " ")

' remove all variables
  {%modskeleton} = @wdrop({%modskeleton},%vnames)

' remove all words that are single digit numbers
  %tmp = "0 1 2 3 4 5 6 7 8 9"
  {%modskeleton} = @wdrop({%modskeleton},%tmp)

' remove all words that are special functions
  %tmp = "exp fact log abs sqrt"
  {%modskeleton} = @wdrop({%modskeleton},%tmp)

  {%modskeleton} = @wunique({%modskeleton})

' remove all remaining words that are numbers
  %tempskel = " "
  for !i9=1 to @wcount({%modskeleton})
    %tmp = @word({%modskeleton},!i9)
    !tmp = @val(%tmp)
    if !tmp = na then
      %tempskel = %tempskel + " " + %tmp
      endif
    next
  {%modskeleton} = %tempskel

  endsub

'************************************************************
'************************************************************
'************************************************************
subroutine replace_operators

' operates on the string %z1

' replace all operators with spaces
  %z1 = @replace(%z1, "+", " ")
  %z1 = @replace(%z1, "-", " ")
  %z1 = @replace(%z1, "*", " ")
  %z1 = @replace(%z1, "/", " ")
  %z1 = @replace(%z1, "^", " ")
  %z1 = @replace(%z1, "=", " ")
  %z1 = @replace(%z1, "(", " ")
  %z1 = @replace(%z1, ")", " ")

  endsub
